/*
 * COPYRIGHT AND LICENSE
 * 
 * Copyright (c) 2004-2005, Juniper Networks, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *       1.      Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the following
 * disclaimer.
 *       2.      Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *       3.      The name of the copyright owner may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <paths.h>
#include <libgen.h>
#include <event.h>

#include "amt.h"
#include "gw.h"

static const char __attribute__((unused)) id[] = "@(#) $Id: gw_main.c,v 1.24 2005/05/02 20:33:32 pusateri Exp $";

static void
usage(char *name) 
{
    fprintf(stderr, "usage: %s -a anycast prefix/plen [-d] [-n]\n", name);
    exit(1);
}

/*
 * Only allow one instance of the AMT gateway to run at a time.
 * Use the PID lockfile method to detect another instance.
 */
static int
init_pid_lockfile(gw_t *gw)
{
    char pidpath[PATH_MAX];
    FILE *fp;

    snprintf(pidpath, MAXPATHLEN, "%s/%s.pid", _PATH_VARRUN, gw->name);
    fp = fopen(pidpath, "r");
    if (fp) {
	char pid_string[11];

	/* PID lock file exists */
	if (fgets(pid_string, sizeof(pid_string), fp) != NULL) {
	    int oldpid;

	    oldpid = (int) strtol(pid_string, (char **)NULL, 10);
	    if (oldpid) {
		if (kill(oldpid, 0) == 0) {
		    /* gateway process already running */
		    return oldpid;
		}
	    }
	}
	fclose(fp);
	unlink(pidpath);
    }

    /*
     * create new pid lock file
     */
    gw->pid = getpid();
    fp = fopen(pidpath, "w");
    if (fp == NULL) {
	return -1;
    }
    
    fprintf(fp, "%d\n", gw->pid);
    fclose(fp);

    return 0;
}

static void
gw_reconfig(int sig)
{
/* XXX */
}

static void
gw_shutdown(int sig)
{
/* XXX */
}

static void
init_signal_handler(gw_t *gw)
{
    if (signal(SIGHUP, gw_reconfig) == SIG_ERR) {
	fprintf(stderr, "%s: couldn't install SIGHUP handler\n", gw->name);
	exit(1);
    }

    if (signal(SIGTERM, gw_shutdown) == SIG_ERR) {
	fprintf(stderr, "%s: couldn't install SIGTERM handler\n", gw->name);
	exit(1);
    }
}

int
main (int argc, char **argv)
{
    int ch, rc, parent, nofork = FALSE;
    gw_t gw;
    struct timeval tv;

    bzero(&gw, sizeof(gw_t));
    gw.anycast_subnet_prefix = 0;
    gw.anycast_subnet_plen = 0;

    while ((ch = getopt(argc, argv, "u:a:dn")) != EOF) {
	char *pstr = NULL;
	in_addr_t prefix;

    switch (ch) {
        case 'u':
            gw.use_unicast_addr = 1;
	case 'a':
	    pstr = strsep(&optarg, "/");
	    if (pstr == NULL) {
		fprintf(stderr, "bad anycast prefix\n");
		exit(1);
	    }
	    if (optarg == NULL) {
		fprintf(stderr, "bad anycast prefix length\n");
		exit(1);
	    }
	    rc = inet_pton(AF_INET, pstr, &prefix);
	    if (rc == 1) {
		gw.anycast_subnet_prefix = ntohl(prefix);
		gw.anycast_subnet_plen = strtol(optarg, NULL, 10);
		if (gw.anycast_subnet_plen == 0) {
		    fprintf(stderr, "bad anycast prefix length\n");
		    exit(1);
		}
		if (gw.anycast_subnet_plen > AMT_HOST_PLEN) {
		    fprintf(stderr, "anycast prefix length too long\n");
		    exit(1);
		}
	    } else {
		fprintf(stderr, "bad anycast prefix\n");
		exit(1);
	    }
            break;

        case 'd':
            gw.debug = TRUE;
            break;

        case 'n':
            nofork = TRUE;
            break;
        default:
        case '?':
            usage(argv[0]);
        }
    }

    /*
     * save name
     */
    snprintf(gw.name, sizeof(gw.name), "%s", basename(argv[0]));

    if (gw.anycast_subnet_plen == 0) {
	fprintf(stderr, "%s: anycast prefix/len must be set with -a\n",
		gw.name);
	exit(1);
    }

    rc = init_pid_lockfile(&gw);
    if (rc < 0) {
	fprintf(stderr, "%s: Can't create pid lockfile.\n", gw.name);
	exit(1);
    } else if (rc) {
	/* already running */
	fprintf(stderr, "%s: already running, pid %d\n", gw.name, rc);
	exit(1);
    }
    
    if (nofork == FALSE) {
	parent = fork();
	if (parent < 0) {
	    fprintf(stderr, "%s: Unable to fork background process.\n",
		    gw.name);
	    exit(1);
	}
	if (parent) { /* parent */
	    exit(0);
	}
    }

    /* child or foreground */

    chdir(_PATH_VARTMP);

    gw.gw_context = event_init();
    if (gw.gw_context == NULL) {
	fprintf(stderr, "event_init failed\n");
	exit(1);
    }

    if (init_iftun_device(&gw) < 0) {
	fprintf(stderr, "%s: Couldn't open tunnel device for writing.\n",
		gw.name);
	exit(1);
    }

    init_signal_handler(&gw);

    rc = init_sockets(&gw);
    if (rc) {
	fprintf(stderr, "%s: Trouble opening sockets: %s.\n", gw.name,
		strerror(rc));
	exit(1);
    }

    /*
     * read interface configuration and clean up any existing anycast
     * addresses on the tunnel interface.
     * Determine a local source address to use.
     */
    rc = init_address(&gw);
    if (rc) {
	exit(1);
    }

    /*
     * Send periodic discoveries until we receive a relay advertisement
     */
    timerclear(&tv);
    tv.tv_sec = GW_DISCOVERY_OFFSET;
    evtimer_set(&gw.discovery_timer, gw_send_discovery, &gw);
    rc = evtimer_add(&gw.discovery_timer, &tv);
    if (rc < 0) {
	fprintf(stderr, "%s: can't initialize discovery timer: %s\n", gw.name,
		strerror(errno));
	exit(1);
    }

    TAILQ_INIT(&gw.request_head);

    rc = event_dispatch();
    fprintf(stderr, "%s: Unexpected exit: %s\n", gw.name, strerror(errno));

    exit(rc);
}

/*
 * we lost communication with the relay
 * flush all pending transactions
 * and try to locate another relay
 * XXX
 */
static void
gw_rediscover_relay(gw_t *gw)
{    
    int rc;
    request_t *rq;
    struct timeval tv;

    /*
     * flush relay state
     */
    gw_cleanup_udp_sock(gw);

    /*
     * close relay socket
     */
    rc = event_del(&gw->udp_event_id);
    if (rc < 0) {
    }
    bzero(&gw->udp_event_id, sizeof(struct event));
    close(gw->udp_sock);
    gw->udp_sock = 0;

    /*
     * Delete tunnel interface
     */
    rc = event_del(&gw->tun_event_id);
    if (rc < 0) {
    }
    bzero(&gw->tun_event_id, sizeof(struct event));
    close(gw->tundev);
    gw->tundev = 0;
    
    /*
     * stop all the request timers
     */
    TAILQ_FOREACH(rq, &gw->request_head, rq_next) {
	if (evtimer_pending(&rq->rq_timer, NULL)) {
	    evtimer_del(&rq->rq_timer);
	}
    }

    /*
     * stop the discovery timer
     */
    if (evtimer_pending(&gw->discovery_timer, NULL)) {
	evtimer_del(&gw->discovery_timer);
    }

    /*
     * stop the query timer
     */
    if (evtimer_pending(&gw->query_timer, NULL)) {
	evtimer_del(&gw->query_timer);
    }

    /*
     * Create tunnel interface
     */
    if (init_iftun_device(gw) < 0) {
	fprintf(stderr, "%s: Couldn't open tunnel device for writing.\n",
		gw->name);
	exit(1);
    }
    /*
     * open discovery socket
     */
    rc = init_sockets(gw);
    if (rc) {
	fprintf(stderr, "%s: Trouble opening sockets: %s.\n", gw->name,
		strerror(rc));
	exit(1);
    }

    /*
     * read interface configuration and clean up any existing anycast
     * addresses on the tunnel interface.
     * Determine a local source address to use.
     */
    rc = init_address(gw);
    if (rc) {
	exit(1);
    }

    /*
     * Send periodic discoveries until we receive a relay advertisement
     */
    timerclear(&tv);
    tv.tv_sec = GW_DISCOVERY_OFFSET;
    evtimer_set(&gw->discovery_timer, gw_send_discovery, gw);
    rc = evtimer_add(&gw->discovery_timer, &tv);
    if (rc < 0) {
	fprintf(stderr, "%s: can't initialize discovery timer: %s\n", gw->name,
		strerror(errno));
	exit(1);
    }
}

/*
 * We're having trouble communicating with the relay.
 * This may be a temporary situation so let's not search for
 * a new relay until some time has passed and we're sure.
 */
void
gw_age_relay(gw_t *gw)
{
/* XXX */
    gw_rediscover_relay(gw);
}

